# include <stdlib.h>
# include <assert.h>
# include <values.h>
# include "gcc-externs.h"
# include "list.h"
# include "sdg.h"
# include "bb.h"
# include "ud.h"
# include "regset.ic"

# define SLICE_PARAMS_FILE   "params"
# define O_LDST_TO_COMP_COPY 4.0
# define O_COMP_TO_LDST_COPY 4.0
# define O_DUPL		         2.0
# define BRANCH_LOSS_FACTOR  1.0

static float o_LdstCompCopy   = O_LDST_TO_COMP_COPY;
static float o_CompLdStCopy   = O_COMP_TO_LDST_COPY;
static float o_dupl           = O_DUPL;
static float branchLossFactor = BRANCH_LOSS_FACTOR;
static int   shouldDuplicate  = 1;
static int   slicingParametersRead = 0;

List sliceNodeList; 	/* The list of nodes from which slicing begins */
List ldstNodeList; 		/* The list of LdSt nodes */
List compNodeList;		/* The list of Comp nodes */
List copyNodeList;		/* The list of LdSt nodes for which copies are needed */

extern rtx GetReplacementReg(rtx);
extern rtx DEST_REG(rtx insn);

/* ##############################################################
 * First phase of the algorithm.  The LdSt slice of the graph is
 * constructed by this part of the code.
 * ############################################################## */

static void
FindBackwardSlice(SdgNode *node)
{
/* Slice the graph backward with respect to the sources of 'node'
 * Since we start off this process with load/store nodes, the resulting
 * slice will be part of the ld-st subsystem */

	Iterator  iter;
	SdgNode  *src;

	assert(!node->visited);

	node->inLdSt = 1;
	node->visited = 1;

		/* For each of the sources, slice backwards with respect to the src.
		 * Finds the backward slice. */
	for (src = (SdgNode *)InitIter(iter, node->srcs);
		 src != NULL; 
		 src = (SdgNode *)GetNext(iter))
	{
		if (!src->visited) {
			FindBackwardSlice(src);
			Append(&ldstNodeList, src);
		}
	}
}

static void
ClearAllVisitedBits()
{
	SdgNode *node;

	for (node = sdg.head; node != NULL; node = node->next)
		node->visited = 0;
}

static void
ClearVisitedBitsInSlice(SdgNode *node)
{
	Iterator  iter;
	SdgNode  *tmp;

	node->visited = 0;
	if (!node->inLdSt) {
		for (tmp = (SdgNode *)InitIter(iter, node->srcs);
			 tmp != NULL;
			 tmp = (SdgNode *)GetNext(iter))
		{
			if (tmp->visited)
				ClearVisitedBitsInSlice(tmp);
		}
	}

	for (tmp = (SdgNode *)InitIter(iter, node->dsts);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (tmp->visited && !tmp->inLdSt)
			ClearVisitedBitsInSlice(tmp);
	}
}

static void
FindAddressComputationSlice()
{
	SdgNode  *node;
	Iterator  iter;
	int       size;

	Init(ldstNodeList);

	for (node = (SdgNode *)RemoveHead(&sliceNodeList);
		 node != NULL;
		 node = (SdgNode *)RemoveHead(&sliceNodeList))
	{
		if (!node->visited) {
			FindBackwardSlice(node);
			Append(&ldstNodeList, node);
		}
	}
}

static void
ClearVisitedBitsInNodeList(List list)
{
	Iterator  iter;
	SdgNode  *tmp;

	tmp = (SdgNode *)InitIter(iter, list);
	while (tmp != NULL) {
		tmp->visited = 0;
		tmp = (SdgNode *)GetNext(iter);
	}
}

/* ##############################################################
 * This part of code actually introduces the copy and duplicate 
 * instructions in the graph.
 * ############################################################## */

void
SubstituteReplacementRegs(rtx copy)
{
/* 'copy' is going to be in Comp.  'Association' contains 
 * register replacements for instrs in Comp.  Go ahead and
 * do all replacements in 'copy'.  This can save some copies */

	Iterator     iter;
	Association *assoc;
	extern List  assocList;

	for (assoc = (Association *)InitIter(iter, assocList);
		 assoc != NULL;
		 assoc = (Association *)GetNext(iter))
	{
		if (assoc->newReg != DEST_REG(copy))
			ReplaceSourceRegInInstr(copy, assoc->oldReg, assoc->newReg);
	}
}

static void
ReplaceRegInDests(SdgNode *node, rtx oldReg, rtx newReg)
{
	extern void ReplaceRegInInstr(rtx insn, rtx oldReg, rtx newReg);

	Iterator  iter, tmpIter;
	SdgNode  *child, *tmp;

	for (child = (SdgNode *)InitIter(iter, node->dsts);
		 child != NULL;
		 child = (SdgNode *)GetNext(iter))
	{
		if (!child->inLdSt) {
			ReplaceSourceRegInInstr(child->instr, oldReg, newReg);
				/*
				 * 'child' might be receiving a value of 'oldReg' from
				 * multiple instrs.  Have to replace 'oldReg' by 'newReg'
				 * in all those instructions.
				 */
			for (tmp = (SdgNode *)InitIter(tmpIter, child->srcs);
				 tmp != NULL;
				 tmp = (SdgNode *)GetNext(tmpIter))
			{
				if (   !tmp->copiesAreNeeded
					&& reg_mentioned_p(oldReg, DEST_REG(tmp->instr)))
				{
					ReplaceDestRegInInstr(tmp->instr, oldReg, newReg);
					ReplaceRegInDests(tmp, oldReg, newReg);
				}
			}
		}
	}
}

void
InsertCopy(SdgNode *node)
{
	SdgNode *copy;
	rtx      src, dest, destCopy;
	rtx      oldReg, newReg;
	rtx      instrCopy;

	extern void add_insn_after(rtx, rtx);

		/* HACK!  This is to prevent 'gcc' from doing a copy
		 * propagation of the copy.  For a symbol-ref, this shouldn't
		 * be done since though symbol-ref is a constant, it uses a
		 * temporary integer reg to compute the address */
	src  = (GET_CODE(node->instr) == SET) ? SET_SRC(node->instr)
										  : SET_SRC(PATTERN(node->instr));
	if (GET_CODE(src) == SYMBOL_REF)
		src->call = 1;

	dest = (GET_CODE(node->instr) == SET) ? SET_DEST(node->instr)
									      : SET_DEST(PATTERN(node->instr));
	destCopy = copy_rtx(dest);

	oldReg = DEST_REG(node->instr);
	newReg = node->replacementReg = GetReplacementReg(oldReg);
	assert(newReg != NULL);
	fprintf(stderr, "Inserting a copy for NODE %d : Copying %d into %d\n", 
			node->id, REGNO(oldReg), REGNO(newReg));
	instrCopy = make_insn_raw(gen_rtx(SET, GET_MODE(newReg), destCopy, dest));
	add_insn_after(instrCopy, (node->inCompoundRtx) ? node->compoundRtx
							  					    : node->instr);
	ReplaceDestRegInInstr(instrCopy, oldReg, newReg);

	SubstituteReplacementRegs(instrCopy);

		/* Insert 'copy' after the load in the graph.
		 * Dont bother to adjust edges */
	copy = AllocNewSDGnode(instrCopy);
	copy->inLdSt = 1;
	copy->next = node->next;
	node->next = copy;
	if (sdg.tail == node)
		sdg.tail = copy;

	ReplaceRegInDests(node, oldReg, newReg);
}

static void
DuplicateInstruction(SdgNode *node)
{
	SdgNode  *duplicate;
	rtx       patCopy, instrCopy;
	rtx       oldReg, newReg;
	Iterator  iter;
	SdgNode  *src;
	extern void add_insn_after(rtx, rtx);

	assert(node->inLdSt);

	patCopy = (GET_CODE(node->instr) == SET) ? copy_rtx(node->instr)
											 : copy_rtx(PATTERN(node->instr));
	instrCopy = make_insn_raw(patCopy);
	add_insn_after(instrCopy, (node->inCompoundRtx) ? node->compoundRtx
												    : node->instr);
	SubstituteReplacementRegs(instrCopy);

	oldReg = DEST_REG(node->instr);
	newReg = node->replacementReg = GetReplacementReg(oldReg);
	assert(newReg != NULL);
	ReplaceDestRegInInstr(instrCopy, oldReg, newReg);
	fprintf(stderr, "DUPLICATING instr(NODE %d) : Copying %d into %d\n", 
				node->id, REGNO(oldReg), REGNO(newReg));

	for (src = (SdgNode *)InitIter(iter, node->srcs);
		 src != NULL;
		 src = (SdgNode *)GetNext(iter))
	{
		if ((src != node) && (src->id >= 4)) {
			if (src->replacementReg == NULL) {
				if (src->shouldDuplicate)
					DuplicateInstruction(src);
				else
					InsertCopy(src);
			}
			oldReg = DEST_REG(src->instr);
			newReg = src->replacementReg;
			ReplaceSourceRegInInstr(instrCopy, oldReg, newReg);
		}
	}

		/* Insert 'duplicate' in the graph. Dont bother to adjust edges */
	duplicate = AllocNewSDGnode(instrCopy);
	duplicate->next = node->next;
	node->next = duplicate;
	if (sdg.tail == duplicate)
		sdg.tail = duplicate;

	ReplaceRegInDests(node, DEST_REG(node->instr), node->replacementReg);
}

# ifdef __GNUC__
inline
# endif
int
HasOnlyOneDst(SdgNode *node)
{
	return (node->dsts.head == node->dsts.tail);
}

static void
IntroduceCopyInstructions()
{
	SdgNode  *node;

	for (node = (SdgNode *)RemoveHead(&copyNodeList);
		 node != NULL;
		 node = (SdgNode *)RemoveHead(&copyNodeList))
	{
			/* (i) Node is a dummy arg node
			 * (ii) An arg. is being produced & 'node' is in Comp
			 * (iii) A ret. val is being assigned & 'node' is in Comp
			 * For (i), do nothing.
			 * gcc automatically handles cases (ii) & (iii).
			 */
		if (   (node->dstUsesIntArgReg && (node->id < 4))
			|| (node->dstUsesIntArgReg && !node->inLdSt)
			|| (node->dstUsesIntRetReg && (!node->inLdSt || HasOnlyOneDst(node))))
			continue;

		if (node->copiesAreNeeded && (node->replacementReg == NULL)) {
			if (node->shouldDuplicate)
				DuplicateInstruction(node);
			else
				InsertCopy(node);
		}
	}
}

/* ##############################################################
 * Second phase of the algorithm.  Builds the boundary of the
 * LdSt slice.  'Boundary' consists of those nodes that are in
 * LdSt but have children not in LdSt.  After the boundary is
 * constructed, boundary expansion is attempted for every non-LdSt
 * child of a boundary node.
 * ############################################################## */

static int
CheckIfCopiesAreNeeded(SdgNode *node)
{
/* Detect those cases where copies are needed.
 * 'node' might be in ldst or comp.  Nothing known.
 */

	Iterator  iter;
	SdgNode  *tmp;

	if (node->inLdSt) {
		for (tmp = (SdgNode *)InitIter(iter, node->dsts);
			 tmp != NULL;
			 tmp = (SdgNode *)GetNext(iter))
		{
			if (!tmp->inLdSt)
				return 1;
		}
	}
# ifdef DEBUG_SLICER
	else {
			/* SANITY CHECK */
		for (tmp = (SdgNode *)InitIter(iter, node->dsts);
			 tmp != NULL;
			 tmp = (SdgNode *)GetNext(iter))
		{
			assert(!tmp->inLdSt);
		}
	}
# endif

	return 0;
}

static float
ComputeDuplicatingCost(SdgNode *node)
{
	Iterator  iter;
	SdgNode  *tmp;
	float     duplCost;

		/* Reflects the overhead of an extra instruction 
		 * that is introduced */
	duplCost = o_dupl * node->nodeWeight;
	for (tmp = (SdgNode *)InitIter(iter, node->srcs);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
# ifdef SIMPLE_HEURISTIC
		if (tmp != node)
			duplCost += tmp->copyingCost;
# elif NEW_DUPL_HEURISTIC
		if (tmp != node && !tmp->copiesAreNeeded)
			duplCost += (tmp->shouldDuplicate) ? tmp->duplicatingCost
											   : tmp->copyingCost;
# else
error define either "SIMPLE_HEURISTIC" or "NEW_DUPL_HEURISTIC"
# endif
	}

	return duplCost;
}

List inLdstList;
List maybeInLdstList;

static int
MightHaveToIntroduceCopies(SdgNode *node)
{
/* GIVEN : 'node->mightBeInLdst' && '!node->inLdSt' */

	Iterator  iter;
	SdgNode  *tmp;

	for (tmp = (SdgNode *)InitIter(iter, node->dsts);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
# ifdef DEBUG_SLICER
		assert(!tmp->inLdSt);
# endif
		if (!tmp->mightBeInLdst)
			return 1;
	}

	return 0;
}

static int
CopiesMayBeNeeded(SdgNode *node)
{
/* GIVEN : 'node->inLdSt'.  
 * Also, one of its children 'mightBeInLdst'. => need not worry if
 * 'node' is a load op.  It is still supplying a node in ldst.
 * So, if there is at least 1 child who is in Comp, or if there is
 * a ldst child who might get duplicated, then copies are needed */

	Iterator  iter;
	SdgNode  *tmp;

	for (tmp = (SdgNode *)InitIter(iter, node->dsts);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (!tmp->inLdSt && !tmp->mightBeInLdst)
			return 1;

		if (tmp != node && tmp->copiesMightBeNeeded && tmp->shouldDuplicate)
			return 1;
	}

	return 0;
}

static void
SeparateNodesInBackwardSlice(SdgNode *node)
{
/* GIVEN : '!node->inLdSt' */

	Iterator  iter;
	SdgNode  *tmp;

	node->visited = 1;
	node->mightBeInLdst = 1;
	Append(&maybeInLdstList, node);

	for (tmp = (SdgNode *)InitIter(iter, node->srcs);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (!tmp->visited) {
			if (tmp->inLdSt)
				Append(&inLdstList, tmp);
			else
				SeparateNodesInBackwardSlice(tmp);
		}
	}
}

static float 
LossInMovingBackwardSliceToLdst(SdgNode *startNode)
{
	Iterator  iter;
	float     loss, tmpLoss;
	int       changed;
	int       numIters;
	SdgNode  *node;

	Init(inLdstList);
	Init(maybeInLdstList);

	SeparateNodesInBackwardSlice(startNode);
	loss = 0.0;
	for (node = (SdgNode *)InitIter(iter, maybeInLdstList);
		 node != NULL;
		 node = (SdgNode *)GetNext(iter))
	{
			/* Since 'node' doesn't execute in 'Comp', loss = nodeWeight.
			 * However, store value nodes correspond to store instructions
			 * and dont incur any loss since they themselves execute in
			 * the LdSt subsystem.
			 * It is good to execute branches in Comp.  So, magnify
			 * the loss for branches */
		if (GET_CODE(node->instr) == JUMP_INSN)
			loss += node->nodeWeight * branchLossFactor;
		else if (!node->isStoreVal && !node->isLoadOp)
			loss += node->nodeWeight;

			/* If this instr. produces an argument value or return value,
			 * it would have to go in a hard reg.  By keeping instr in 'LdSt'
			 * I dont have to do that copy.  So, that is a profit */
		if (node->dstUsesIntArgReg || node->dstUsesIntRetReg)
			loss -= (o_CompLdStCopy * node->nodeWeight);

			/* Also, now if I have any children in Comp, I would have to
			 * introduce copies.  So, that is an additional loss. */
		if (MightHaveToIntroduceCopies(node)) {
			node->copiesMightBeNeeded = 1;
			loss += (node->shouldDuplicate) ? node->duplicatingCost
											: node->copyingCost;
		}
			/* For nodes that produce argument/return values,
			 * copies are needed as long as they are in Comp */
		else if (!node->dstUsesIntArgReg && !node->dstUsesIntRetReg)
			node->copiesMightBeNeeded = 0;
	}

	numIters = 0;
	do {
			/* Without duplication, iteration is not necessary.
			 * With duplicates, copies would need to be introduced because
			 * ldst nodes get duplicated forcing copies for parents */
		changed = 0;
		tmpLoss = 0.0;
		numIters++;
		for (node = (SdgNode *)InitIter(iter, inLdstList);
			 node != NULL;
			 node = (SdgNode *)GetNext(iter))
		{
			int copiesNeededNow = CopiesMayBeNeeded(node);
			if (!node->copiesMightBeNeeded && copiesNeededNow) {
				changed = 1;
				node->copiesMightBeNeeded = 1;
				tmpLoss += (node->shouldDuplicate ? node->duplicatingCost
											      : node->copyingCost);
			}
			else if (node->copiesMightBeNeeded && !copiesNeededNow) {
				changed = 1;
				node->copiesMightBeNeeded = 0;
				tmpLoss -= (node->shouldDuplicate ? node->duplicatingCost
											      : node->copyingCost);
			}
		}
	} while (changed);

	if (numIters > 2)
		fprintf(stderr, "# iterations to converge - %d\n", numIters);

	ClearVisitedBitsInNodeList(maybeInLdstList);
	ClearVisitedBitsInNodeList(inLdstList);

	FreeList(&inLdstList);
	FreeList(&maybeInLdstList);

	return (loss + tmpLoss);
}

static void
PutChildrenIntoCompList(SdgNode *node)
{
	SdgNode  *child;
	Iterator  iter;

	for (child = (SdgNode *)InitIter(iter, node->dsts);
		 child != NULL;
		 child = (SdgNode *)GetNext(iter))
	{
		if (!child->inLdSt && (child != node)) {
			if (!child->insertedInList) {
				Append(&compNodeList, child);
				child->insertedInList = 1;
			}
		}
	}
}

# ifdef __GNUC__
inline
# endif
static int
AllowedToDuplicateNode(SdgNode *node)
{
		/* Loads/stores cannot be duplicated
		 * Call nodes cannot be duplicated
		 * Instrs using hard int regs cannot be duplicated
		 * Instrs in compound rtx cannot be duplicated
		 * Dummy arg. nodes cannot be duplicated
		 * (But no need to check for dummy arg nodes. Wont call this function 
		 *  for dummy argument nodes at all because they are not in the SDG
		 *  and because they are in LdSt)
		 */
	return !(   node->isMemOp 
			 || node->isACall
			 || node->srcUsesHardIntReg 
			 || node->dstUsesHardIntReg
			 || node->inCompoundRtx);
}

static void
ReexamineChildren(SdgNode *node)
{
	SdgNode  *child;
	Iterator  iter;

	for (child = (SdgNode *)InitIter(iter, node->dsts);
		 child != NULL;
		 child = (SdgNode *)GetNext(iter))
	{
		if (!child->inLdSt && (child != node)) {
# ifdef NEW_DUPL_HEURISTIC
				/* Recompute duplicating cost and 
				 * see if child can be duplicated */
			if (shouldDuplicate && AllowedToDuplicateNode(child)) {
				child->duplicatingCost = ComputeDuplicatingCost(child);
				child->shouldDuplicate = (child->duplicatingCost < child->copyingCost);
			}
# endif
			if (!child->insertedInList) {
				Append(&compNodeList, child);
				child->insertedInList = 1;
			}
		}
	}
}

static void
MoveBackwardSliceToLdst(SdgNode *node)
{
	Iterator  iter;
	SdgNode  *tmp;

	node->copiesAreNeeded = node->copiesMightBeNeeded;
	if (!node->inLdSt) {
			/* Push 'node' into 'LdSt' */
		node->mightBeInLdst = 0;
		node->inLdSt = 1;
		Append(&ldstNodeList, node);

			/* Push all sources into 'LdSt' too */
		for (tmp = (SdgNode *)InitIter(iter, node->srcs);
			 tmp != NULL;
			 tmp = (SdgNode *)GetNext(iter))
		{
			MoveBackwardSliceToLdst(tmp);
		}
	}

		/* Will have to re-examine 'dsts' of 'node'
		 * and the benefit of keeping them in Comp */
	ReexamineChildren(node);
}

void
UnmarkBackwardSlice(SdgNode *node)
{
/* Reset 'mightBeInLdst' and 'copiesMightBeNeeded' */

	Iterator  iter;
	SdgNode  *tmp;

	node->mightBeInLdst = 0;
	node->copiesMightBeNeeded = node->copiesAreNeeded;
	for (tmp = (SdgNode *)InitIter(iter, node->srcs);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (tmp->mightBeInLdst)
			UnmarkBackwardSlice(tmp);
		else if (tmp->inLdSt)
			tmp->copiesMightBeNeeded = tmp->copiesAreNeeded;
	}
}

static void
BuildCompList()
{
	SdgNode  *node;
	Iterator  iter;
	int       i, changed, numIters;

	Init(compNodeList);

		/* Add all children of dummy argument nodes to 'compList'
		 * This allows us to handle copies for arguments in the
		 * same way the rest of the copies are handled */
	for (i = 0; i < GP_ARG_LAST - GP_ARG_FIRST + 1; i++) {
		SdgNode *node = dummyArgNodes[i];
		node->shouldDuplicate = 0;
		node->copyingCost = o_LdstCompCopy * node->nodeWeight;
		node->copiesMightBeNeeded = node->copiesAreNeeded
								  = CheckIfCopiesAreNeeded(node);
		if (node->copiesAreNeeded)
			PutChildrenIntoCompList(dummyArgNodes[i]);
	}

		/* Compute copying cost for all nodes.  This is 
		 * to avoid recomputing it every time it is needed */
	for (node = sdg.head; node != NULL; node = node->next) {
		node->copyingCost = o_LdstCompCopy * node->nodeWeight;

			/* To uniformly handle copies due to argument copying
			 * insert all nodes that produce arguments/return values
			 * into comp node list so that they will get examined
			 * for profit in introducing copies
			 */
		if (   (node->dstUsesIntArgReg || node->dstUsesIntRetReg)
			&& !node->inLdSt)
		{
			node->copiesMightBeNeeded = node->copiesAreNeeded = 1;
/****
			if (!node->insertedInList) {
				Append(&compNodeList, node);
				node->insertedInList = 1;
			}
****/
		}
	}

		/* If copies are needed for a node, put it in the comp list
		 * The node will be examined for profitability for introducing
		 * copies */
	for (node = (SdgNode *)InitIter(iter, ldstNodeList);
		 node != NULL;
		 node = (SdgNode *)GetNext(iter))
	{
		node->copiesMightBeNeeded = node->copiesAreNeeded
								  = CheckIfCopiesAreNeeded(node);
		if (node->copiesAreNeeded)
			PutChildrenIntoCompList(node);
	}

		/* For every node in the graph, compute 'duplicatingCost'.
		 * This is the overhead charged to this node for duplication
		 * if and when necessary.
		 * NOTE : For the current heuristics, 'duplicatingCost' cannot 
		 * be computed before 'copyingCost' is computed
		 */
	if (shouldDuplicate) {
		numIters = 0;
		do {
			changed = 0;
			numIters++;
			for (node = sdg.head; node != NULL; node = node->next) {
				if (AllowedToDuplicateNode(node)) {
					float newCost = ComputeDuplicatingCost(node);
					if (newCost != node->duplicatingCost) {
						changed = 1;
						node->duplicatingCost = newCost;
						node->shouldDuplicate = (node->duplicatingCost < node->copyingCost);
					}
				}
			}
		} while (changed);
	
		fprintf(stderr, "# iters for dupl cost to converge: %d\n", numIters);
	}
}

static void
EliminateDefinitelyLossyNodes()
{
	SdgNode  *node;
	Iterator  iter;

	ClearAllVisitedBits();

	for (node = (SdgNode *)RemoveHead(&compNodeList);
		 node != NULL;
		 node = (SdgNode *)RemoveHead(&compNodeList))
	{
		node->insertedInList = 0;
		if (!node->inLdSt) {
				/* 'loss < 0' is a no-win case for 'Comp'
				 * Just push 'BackwardSlice(node)' into LdSt.
				 * 'loss = 0' is a tricky case.  Examine from children
				 * onwards if loss would be greater.  If so, I might end
				 * up retaining 'node' in Comp after all. */
			float loss = LossInMovingBackwardSliceToLdst(node);
			if (loss < 0)
				MoveBackwardSliceToLdst(node);
			else {
				if (loss == 0)
					PutChildrenIntoCompList(node);

				UnmarkBackwardSlice(node);
			}
		}
	}
}

static void
BuildCopyList()
{
	int      i;
	SdgNode *node;

	Init(copyNodeList);

	for (i = 0; i < GP_ARG_LAST - GP_ARG_FIRST + 1; i++) {
		dummyArgNodes[i]->visited = 0;
		Append(&copyNodeList, dummyArgNodes[i]);
	}

	FreeList(&ldstNodeList);

	for (node = sdg.head; node != NULL; node = node->next) {
		if (node->copiesAreNeeded) {
			node->visited = 0;
			Append(&copyNodeList, node);
		}
	}
}

/* ##############################################################
 * Final phase of the algorithm.  Traverse the graph and compute
 * 'profit' using the cost model.  If 'profit' of a "slice" is
 * negative, move it to LdSt, else retain it in Comp.  There is
 * no attempt to maximize profit if 'Profit > 0'.
 * ############################################################## */

List boundaryNodes;

static float
BenefitOfSlice(SdgNode *node)
{
	Iterator iter;
	SdgNode *src, *dst;
	float    benefit;

	node->visited = 1;
	benefit = 0.0;
	if (node->copiesAreNeeded) {
		if (node->inLdSt && !node->insertedInList) {
			Append(&boundaryNodes, node);
			node->insertedInList = 1;
		}
		else {
			assert(node->dstUsesIntArgReg || node->dstUsesIntRetReg);
			benefit = -node->nodeWeight * o_CompLdStCopy;
		}
	}
	else {
			/* No benefit for 'store/loads'.  They anyway execute in LdSt. */
		if (!node->isStoreVal && !node->isLoadOp)
			benefit += (GET_CODE(node->instr) == JUMP_INSN)
							? node->nodeWeight * branchLossFactor
							: node->nodeWeight;
	}

	if (!node->inLdSt) {
		for (src = (SdgNode *)InitIter(iter, node->srcs);
			 src != NULL;
			 src = (SdgNode *)GetNext(iter))
		{
			if (!src->visited)
				benefit += BenefitOfSlice(src);
		}
	}

	for (dst = (SdgNode *)InitIter(iter, node->dsts);
		 dst != NULL;
		 dst = (SdgNode *)GetNext(iter))
	{
		if (!dst->visited && !dst->inLdSt)
			benefit += BenefitOfSlice(dst);
	}

	return benefit;
}

static float
OverheadOfSlice()
{
	Iterator iter1, iter2;
	SdgNode *node, *src;
	float    overhead = 0.0;

	for (node = (SdgNode *)InitIter(iter1, boundaryNodes);
		 node != NULL;
		 node = (SdgNode *)GetNext(iter1))
	{
		if (node->shouldDuplicate) {
			overhead += node->nodeWeight * o_dupl;

				/* Add parents to the boundary since they need to
				 * be copied/duplicated too! */
			for (src = (SdgNode *)InitIter(iter2, node->srcs);
				 src != NULL;
				 src = (SdgNode *)GetNext(iter2))
			{
				if (!src->insertedInList) {
					src->insertedInList = 1;
					Append(&boundaryNodes, src);
				}
			}
		}
		else
			overhead += node->nodeWeight * o_LdstCompCopy;
	}

	for (node = (SdgNode *)RemoveHead(&boundaryNodes);
	     node != NULL;
	     node = (SdgNode *)RemoveHead(&boundaryNodes))
	{
		node->insertedInList = 0;
	}

	return overhead;
}

static void
MoveSliceIntoLdst(SdgNode *node)
{
	Iterator iter;
	SdgNode *tmp;

	node->copiesAreNeeded = 0;
	node->visited = 1;
	if (!node->inLdSt) {
		node->inLdSt = 1;
		for (tmp = (SdgNode *)InitIter(iter, node->srcs);
			 tmp != NULL;
			 tmp = (SdgNode *)GetNext(iter))
		{
			if (!tmp->visited)
				MoveSliceIntoLdst(tmp);
		}
	}

	for (tmp = (SdgNode *)InitIter(iter, node->dsts);
		 tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (!tmp->visited && !tmp->inLdSt)
			MoveSliceIntoLdst(tmp);
	}
}

static void
EliminateSlicesWithNegativeProfits()
{
	Iterator iter;
	SdgNode *node;
	float    benefit, overhead;

	for (node = (SdgNode *)InitIter(iter, copyNodeList);
		 node != NULL;
		 node = (SdgNode *)GetNext(iter))
	{
		if (node->copiesAreNeeded && !node->visited) {
			Init(boundaryNodes);
			benefit  = BenefitOfSlice(node);
			overhead = OverheadOfSlice();
			FreeList(&boundaryNodes);
			if (benefit < overhead) {
				ClearVisitedBitsInSlice(node);
				MoveSliceIntoLdst(node);
			}
		}
	}
}

/* ##############################################################
 * Read in slicing parameters from the file.  This file specifies
 * 'o_copy' and 'o_dupl' among other things.
 * ############################################################## */

static void
ReadInSlicingParameters()
{
	char  buf[100];
	FILE *fp = fopen(SLICE_PARAMS_FILE, "r");

	if (fp == NULL) {
		perror(SLICE_PARAMS_FILE);
		fprintf(stderr, "Using default values.\n");
		return;
	}

	while (1) {
		fscanf(fp, "%s", buf);
		if (strcmp(buf, "o_ldst_to_comp_copy") == 0)
			fscanf(fp, "%f", &o_LdstCompCopy);
		else if (strcmp(buf, "o_comp_to_ldst_copy") == 0)
			fscanf(fp, "%f", &o_CompLdStCopy);
		else if (strcmp(buf, "o_dupl") == 0)
			fscanf(fp, "%f", &o_dupl);
		else if (strcmp(buf, "branch_loss_factor") == 0)
			fscanf(fp, "%f", &branchLossFactor);
		else if (strcmp(buf, "should_duplicate") == 0)
			fscanf(fp, "%d", &shouldDuplicate);
		else if (strcmp(buf, "end") == 0) {
			fclose(fp);
			return;
		}
		else
			fprintf(stderr, "Unknown parameter %s\n", buf);

		if (feof(fp)) {
			fclose(fp);
			return;
		}
	}
}

void
SliceDependencyGraph()
{
/* 'sliceNodeList' is the list of all load/store instructions
 * AND all nodes which are already determined to be in the load/store slice
 * (instructions which pass parameters to functions, for example) */

	if (!slicingParametersRead) {
		ReadInSlicingParameters();
		slicingParametersRead = 1;
		fprintf(stderr, "Comp to Ldst copy cost : %1.1f\n", o_CompLdStCopy);
		fprintf(stderr, "Ldst to Comp copy cost : %1.1f\n", o_LdstCompCopy);
		fprintf(stderr, "Duplicating cost       : %1.1f\n", o_dupl);
		fprintf(stderr, "Should duplicate ?     : %s\n", (shouldDuplicate ? "yes" : "no"));
	}

		/* Pull all address computation nodes into ld-st system */
	FindAddressComputationSlice();

		/* Build the list of nodes that need to be examined */
	BuildCompList();

		/* Now, try to eliminate nodes from Comp that are definitely lossy */
	EliminateDefinitelyLossyNodes();

		/* Build the list of nodes for which copies are needed */
	BuildCopyList();

		/* Clear all visited bits */
	ClearAllVisitedBits();

		/* Now, try to eliminate slices that are not profitable */
	EliminateSlicesWithNegativeProfits();

		/* Now introduce copies/duplicate code as required */
	IntroduceCopyInstructions();
}
